package edu.upv_ehu.gbg.docencia.perfectos;

import java.math.BigInteger;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

/**
 * Proporciona 5 alternativas para mostrar los números perfectos hasta un tope dado.
 * De utilidad para ver cómo el planteamiento de algoritmos adecuados es importante,
 * y para ver las posibilidades que da Java para realizar cómputos complejos.
 * 
 * @author german
 */
public class CalculadorDiversoDeNumerosPerfectos {
    private CalculadorDiversoDeNumerosPerfectos(){}
    
    //tiempo inicial para mostrar el acumulado hasta la obtención de cada perfecto.
    //en los métodos BIG
    private static long elapsedTimeControl;
    
//----------   METODO NAIF    -------------------------------------------
    
   /**
     * Lleva a la salida estándar los perfectos hasta "max",
     * probando para cada uno de ellos todos sus divisores.
     * 
     * @param max número superior a considerar
     */
    public static void perfectosNaif(long max){
    for (long candidato=1; candidato<=max; candidato++) {
        long suma=0;
        for (long divisor=1 ;divisor<candidato;divisor++) if (candidato%divisor==0) suma+=divisor;        
        if (candidato==suma) System.out.println(candidato);
    }
}
    
    //----------   METODO MEJOR   -------------------------------------------
    
   /**
     * Lleva a la salida estándar los perfectos hasta "max",
     * probando para cada uno los divisores hasta su raíz cuadrada y sumando los encontrados y sus complementarios.
     * 
     * @param max número superior a considerar
     */
    public static void perfectosMejor(long max){
    for (long candidato=1; candidato<max; candidato++) {
        long suma=1;
        for (int divisor=2;divisor<=Math.sqrt(candidato);divisor++) if (candidato%divisor==0) suma+=divisor+candidato/divisor;        
        if (candidato==suma) System.out.println(candidato);
    }
}
    
    //----------   METODO BUENO    -------------------------------------------

   /**
     * Calcula perfectos mediante su expresión basada en primos de Mersenne (2<sup>&rho;-1</sup>)*(2<sup>&rho;</sup>-1). 
     * 
     * @param maxRo valor máximo de "p" en la fórmula a calcular 
     */
    public static void perfectosBueno(int maxRo){
        long ro=2, terminoA=2, terminoB=3;
        Intentos:
        for (; ro<=maxRo; terminoA<<=1, terminoB=(terminoB<<1) | 1L, ro++) {
            for (int divisor=2; divisor<=Math.sqrt(terminoB);divisor++) if (terminoB%divisor==0) continue Intentos;
             System.out.println(terminoA*terminoB);
        }
}
    
    //----------   METODO BIG    ---------------------------------------------

  /**
     * Calcula perfectos mediante su expresión basada en primos de Mersenne (2<sup>&rho;-1</sup>)*(2<sup>&rho;</sup>-1),
     * utilizando una representación numérica de precisión indefinida. 
     * OJO. El resultado no es seguro. La primalidad es probabilística.
     * Para cada uno presenta: tiempo de ejecución acumulado (ms.), p (potencia de Mesrenne), número de cifras del perfecto, el perfecto
     * 
     * @param maxRo máxima potencia (p en la fórmula)
     * @param certainty factor de probabilidad del test de primalidad según err&lt;=(1-1/2^factProb)
     */
    public static void perfectosBig(int maxRo,int certainty){
        System.out.println("Certeza de primalidad: 1-1/2^"+certainty+" = "+(1-1.0/(1<<certainty))); 
        System.out.println("       t(seg.)      ro  long. perfecto");
        String perfecto;
        int ro=2; 
        BigInteger terminoA=new BigInteger("2"), terminoB=new BigInteger("3");
        elapsedTimeControl=System.currentTimeMillis();
        for (;ro<=maxRo;ro++, terminoA=terminoA.shiftLeft(1), terminoB=terminoB.shiftLeft(1).setBit(0))
            if (terminoB.isProbablePrime(certainty))
                System.out.printf("%10.3f seg. %6d %6d %s\n",(System.currentTimeMillis()-elapsedTimeControl)/1000.0,ro,(perfecto=terminoA.multiply(terminoB).toString()).length(),perfecto);
        
    }
    
    //----------   METODO BIGTHREADED    --------------------------------------
   
    /**
     * Calcula perfectos mediante su expresión basada en primos de Mersenne (2<sup>&rho;-1</sup>)*(2<sup>&rho;</sup>-1),
     * utilizando una representación numérica de precisión indefinida y usando todos los hilos del procesador. 
     * OJO. El resultado no es seguro. La primalidad es probabilística.
     * La evaluación de cada candidato se delega en hilos "DeterminaPerfecto" que son ejecutados por un ThreadPool.
     * 
     * @param maxRo máxima potencia (p)
     * @param certainty certeza del test de primalidad según err&lt;=(1-1/2^certainty)
     * @param numeroDeHilos numero de hilos a utilizar (aparte del main)
     */
    public static void perfectosBigThreaded(int maxRo, int certainty,int numeroDeHilos) {
        System.out.println("Ejecutando con "+numeroDeHilos+" hilos");
        System.out.println("Certeza de primalidad: 1-1/2^"+certainty+" = "+(1-1.0/(1<<certainty))); 
        System.out.println("       t(seg.)      ro  long. perfecto");
        ThreadPoolExecutor executor =  new ThreadPoolExecutor(numeroDeHilos, numeroDeHilos, 1, TimeUnit.DAYS, new ArrayBlockingQueue<Runnable>(numeroDeHilos+4), new ThreadPoolExecutor.CallerRunsPolicy());
         String perfecto;
        int ro=2;
        BigInteger terminoA=new BigInteger("2"), terminoB=new BigInteger("3");
        elapsedTimeControl=System.currentTimeMillis();
        for (;ro<=maxRo;ro++, terminoA=terminoA.shiftLeft(1), terminoB=terminoB.shiftLeft(1).setBit(0))
            executor.execute(new DeterminaPerfecto(ro, terminoA, terminoB, certainty));
        executor.shutdown();
        try {executor.awaitTermination(1, TimeUnit.DAYS);
        } catch (InterruptedException ex) {System.err.println("Oops! "+ex.getMessage());
        }
    }    
     
   /**
     * Clase auxiliar que lleva la determinación de si un candidato es perfecto a un hilo, para poder lanzar varios candidatos en paralelo. 
     * 
     * Presenta el tiempo de ejecución acumulado (seg.), p (potencia de Mersenne), número de cifras del perfecto, el perfecto
     */
    private static class DeterminaPerfecto implements Runnable {
        private final int ro,certainty;
        private final BigInteger terminoA, terminoB;
        
        public DeterminaPerfecto(int ro, BigInteger terminoA, BigInteger terminoB, int confidence) {
            this.ro=ro;
            this.terminoA = terminoA;
            this.terminoB = terminoB;
            this.certainty=confidence;
        }
        
        @Override
        public void run() {
            Thread.currentThread().setPriority(4);
            String perfecto;
            if (terminoB.isProbablePrime(certainty))
                System.out.printf("%10.3f seg. %6d %6d %s\n",(System.currentTimeMillis()-elapsedTimeControl)/1000.0,ro,(perfecto=terminoA.multiply(terminoB).toString()).length(),perfecto);
        }
        
    }
    
}